"""Length."""
# -*- coding: utf-8 -*-

# local
from .between import between
from .utils import validator


@validator
def length(value: str, /, *, min_val: int = 0, max_val: int = 0):
    """Return whether or not the length of given string is within a specified range.

    Examples:
        >>> length('something', min_val=2)
        # Output: True
        >>> length('something', min_val=9, max_val=9)
        # Output: True
        >>> length('something', max_val=5)
        # Output: ValidationFailure(func=length, ...)

    Args:
        value:
            The string to validate.
        min_val:
            The minimum required length of the string. If not provided,
            minimum length will not be checked.
        max_val:
            The maximum length of the string. If not provided,
            maximum length will not be checked.

    Returns:
        (Literal[True]):
            If `len(value)` is in between the given conditions.
        (ValidationFailure):
            If `len(value)` is not in between the given conditions.

    > *New in version 0.2.0*.
    """
    return between(len(value), min_val=min_val, max_val=max_val)



import re

from .utils import validator


@validator
def card_number(value):
    """
    Return whether or not given value is a valid card number.

    This validator is based on Luhn algorithm.

    .. luhn:
       https://github.com/mmcloughlin/luhn

    Examples::

        >>> card_number('4242424242424242')
        True

        >>> card_number('4242424242424241')
        ValidationFailure(func=card_number, args={'value': '4242424242424241'})

    .. versionadded:: 0.15.0

    :param value: card number string to validate
    """
    try:
        digits = list(map(int, value))
        odd_sum = sum(digits[-1::-2])
        even_sum = sum([sum(divmod(2 * d, 10)) for d in digits[-2::-2]])
        return (odd_sum + even_sum) % 10 == 0
    except ValueError:
        return False


@validator
def visa(value):
    """
    Return whether or not given value is a valid Visa card number.

    Examples::

        >>> visa('4242424242424242')
        True

        >>> visa('2223003122003222')
        ValidationFailure(func=visa, args={'value': '2223003122003222'})

    .. versionadded:: 0.15.0

    :param value: Visa card number string to validate
    """
    pattern = re.compile(r'^4')
    return card_number(value) and len(value) == 16 and pattern.match(value)


@validator
def mastercard(value):
    """
    Return whether or not given value is a valid Mastercard card number.

    Examples::

        >>> mastercard('5555555555554444')
        True

        >>> mastercard('4242424242424242')
        ValidationFailure(func=mastercard, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: Mastercard card number string to validate
    """
    pattern = re.compile(r'^(51|52|53|54|55|22|23|24|25|26|27)')
    return card_number(value) and len(value) == 16 and pattern.match(value)


@validator
def amex(value):
    """
    Return whether or not given value is a valid American Express card number.

    Examples::

        >>> amex('378282246310005')
        True

        >>> amex('4242424242424242')
        ValidationFailure(func=amex, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: American Express card number string to validate
    """
    pattern = re.compile(r'^(34|37)')
    return card_number(value) and len(value) == 15 and pattern.match(value)


@validator
def unionpay(value):
    """
    Return whether or not given value is a valid UnionPay card number.

    Examples::

        >>> unionpay('6200000000000005')
        True

        >>> unionpay('4242424242424242')
        ValidationFailure(func=unionpay, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: UnionPay card number string to validate
    """
    pattern = re.compile(r'^62')
    return card_number(value) and len(value) == 16 and pattern.match(value)


@validator
def diners(value):
    """
    Return whether or not given value is a valid Diners Club card number.

    Examples::

        >>> diners('3056930009020004')
        True

        >>> diners('4242424242424242')
        ValidationFailure(func=diners, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: Diners Club card number string to validate
    """
    pattern = re.compile(r'^(30|36|38|39)')
    return (
        card_number(value) and len(value) in [14, 16] and pattern.match(value)
    )


@validator
def jcb(value):
    """
    Return whether or not given value is a valid JCB card number.

    Examples::

        >>> jcb('3566002020360505')
        True

        >>> jcb('4242424242424242')
        ValidationFailure(func=jcb, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: JCB card number string to validate
    """
    pattern = re.compile(r'^35')
    return card_number(value) and len(value) == 16 and pattern.match(value)


@validator
def discover(value):
    """
    Return whether or not given value is a valid Discover card number.

    Examples::

        >>> discover('6011111111111117')
        True

        >>> discover('4242424242424242')
        ValidationFailure(func=discover, args={'value': '4242424242424242'})

    .. versionadded:: 0.15.0

    :param value: Discover card number string to validate
    """
    pattern = re.compile(r'^(60|64|65)')
    return card_number(value) and len(value) == 16 and pattern.match(value)



from .utils import validator


@validator
def truthy(value):
    """
    Validate that given value is not a falsey value.

    This validator is based on `WTForms DataRequired validator`_.

    .. _WTForms DataRequired validator:
       https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

    Examples::

        >>> truthy(1)
        True

        >>> truthy('someone')
        True

        >>> truthy(0)
        ValidationFailure(func=truthy, args={'value': 0})

        >>> truthy('    ')
        ValidationFailure(func=truthy, args={'value': '    '})

        >>> truthy(False)
        ValidationFailure(func=truthy, args={'value': False})

        >>> truthy(None)
        ValidationFailure(func=truthy, args={'value': None})

    .. versionadded:: 0.2
    """
    return (
        value and
        (not isinstance(value, str) or value.strip())
    )



"""BTC Address."""
# -*- coding: utf-8 -*-

# standard
from hashlib import sha256
import re

# local
from .utils import validator

_segwit_pattern = re.compile(r"^(bc|tc)[0-3][02-9ac-hj-np-z]{14,74}$")


def _decode_base58(addr: str):
    """Decode base58."""
    alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
    return sum((58**enm) * alphabet.index(idx) for enm, idx in enumerate(addr[::-1]))


def _validate_old_btc_address(addr: str):
    """Validate P2PKH and P2SH type address."""
    if len(addr) not in range(25, 35):
        return False
    decoded_bytes = _decode_base58(addr).to_bytes(25, "big")
    header, checksum = decoded_bytes[:-4], decoded_bytes[-4:]
    return checksum == sha256(sha256(header).digest()).digest()[:4]


@validator
def btc_address(value: str, /):
    """Return whether or not given value is a valid bitcoin address.

    Full validation is implemented for P2PKH and P2SH addresses.
    For segwit addresses a regexp is used to provide a reasonable
    estimate on whether the address is valid.

    Examples:
        >>> btc_address('3Cwgr2g7vsi1bXDUkpEnVoRLA9w4FZfC69')
        # Output: True
        >>> btc_address('1BvBMsEYstWetqTFn5Au4m4GFg7xJaNVN2')
        # Output: ValidationFailure(func=btc_address, args=...)

    Args:
        value:
            Bitcoin address string to validate.

    Returns:
        (Literal[True]):
            If `value` is a valid bitcoin address.
        (ValidationFailure):
            If `value` is an invalid bitcoin address.

    > *New in version 0.18.0*.
    """
    if value and type(value) is str:
        return (
            _segwit_pattern.match(value)
            if value[:2] in ("bc", "tb")
            else _validate_old_btc_address(value)
        )
    return False



from .utils import validator


@validator
def ipv4(value):
    """
    Return whether a given value is a valid IP version 4 address.

    This validator is based on `WTForms IPAddress validator`_

    .. _WTForms IPAddress validator:
       https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

    Examples::

        >>> ipv4('123.0.0.7')
        True

        >>> ipv4('900.80.70.11')
        ValidationFailure(func=ipv4, args={'value': '900.80.70.11'})

    .. versionadded:: 0.2

    :param value: IP address string to validate
    """
    groups = value.split(".")
    if (
        len(groups) != 4
        or any(not x.isdigit() for x in groups)
        or any(len(x) > 3 for x in groups)
    ):
        return False
    return all(0 <= int(part) < 256 for part in groups)


@validator
def ipv4_cidr(value):
    """
    Return whether a given value is a valid CIDR-notated IP version 4
    address range.

    This validator is based on RFC4632 3.1.

    Examples::

        >>> ipv4_cidr('1.1.1.1/8')
        True

        >>> ipv4_cidr('1.1.1.1')
        ValidationFailure(func=ipv4_cidr, args={'value': '1.1.1.1'})
    """
    try:
        prefix, suffix = value.split('/', 2)
    except ValueError:
        return False
    if not ipv4(prefix) or not suffix.isdigit():
        return False
    return 0 <= int(suffix) <= 32


@validator
def ipv6(value):
    """
    Return whether a given value is a valid IP version 6 address
    (including IPv4-mapped IPv6 addresses).

    This validator is based on `WTForms IPAddress validator`_.

    .. _WTForms IPAddress validator:
       https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

    Examples::

        >>> ipv6('abcd:ef::42:1')
        True

        >>> ipv6('::ffff:192.0.2.128')
        True

        >>> ipv6('::192.0.2.128')
        True

        >>> ipv6('abc.0.0.1')
        ValidationFailure(func=ipv6, args={'value': 'abc.0.0.1'})

    .. versionadded:: 0.2

    :param value: IP address string to validate
    """
    ipv6_groups = value.split(':')
    if len(ipv6_groups) == 1:
        return False
    ipv4_groups = ipv6_groups[-1].split('.')

    if len(ipv4_groups) > 1:
        if not ipv4(ipv6_groups[-1]):
            return False
        ipv6_groups = ipv6_groups[:-1]
    else:
        ipv4_groups = []

    count_blank = 0
    for part in ipv6_groups:
        if not part:
            count_blank += 1
            continue
        try:
            num = int(part, 16)
        except ValueError:
            return False
        else:
            if not 0 <= num <= 65536 or len(part) > 4:
                return False

    max_groups = 6 if ipv4_groups else 8
    part_count = len(ipv6_groups) - count_blank
    if count_blank == 0 and part_count == max_groups:
        # no :: -> must have size of max_groups
        return True
    elif count_blank == 1 and ipv6_groups[-1] and ipv6_groups[0] and part_count < max_groups:
        # one :: inside the address or prefix or suffix : -> filter least two cases
        return True
    elif count_blank == 2 and part_count < max_groups and (
            ((ipv6_groups[0] and not ipv6_groups[-1]) or (not ipv6_groups[0] and ipv6_groups[-1])) or ipv4_groups):
        # leading or trailing :: or : at end and begin -> filter last case
        # Check if it has ipv4 groups because they get removed from the ipv6_groups
        return True
    elif count_blank == 3 and part_count == 0:
        # :: is the address -> filter everything else
        return True
    return False


@validator
def ipv6_cidr(value):
    """
    Returns whether a given value is a valid CIDR-notated IP version 6
    address range.

    This validator is based on RFC4632 3.1.

    Examples::

        >>> ipv6_cidr('::1/128')
        True

        >>> ipv6_cidr('::1')
        ValidationFailure(func=ipv6_cidr, args={'value': '::1'})
    """
    try:
        prefix, suffix = value.split('/', 2)
    except ValueError:
        return False
    if not ipv6(prefix) or not suffix.isdigit():
        return False
    return 0 <= int(suffix) <= 128



import re

from .utils import validator

regex = (
    r'^[A-Z]{2}[0-9]{2}[A-Z0-9]{11,30}$'
)
pattern = re.compile(regex)


def char_value(char):
    """A=10, B=11, ..., Z=35
    """
    if char.isdigit():
        return int(char)
    else:
        return 10 + ord(char) - ord('A')


def modcheck(value):
    """Check if the value string passes the mod97-test.
    """
    # move country code and check numbers to end
    rearranged = value[4:] + value[:4]
    # convert letters to numbers
    converted = [char_value(char) for char in rearranged]
    # interpret as integer
    integerized = int(''.join([str(i) for i in converted]))
    return (integerized % 97 == 1)


@validator
def iban(value):
    """
    Return whether or not given value is a valid IBAN code.

    If the value is a valid IBAN this function returns ``True``, otherwise
    :class:`~validators.utils.ValidationFailure`.

    Examples::

        >>> iban('DE29100500001061045672')
        True

        >>> iban('123456')
        ValidationFailure(func=iban, ...)

    .. versionadded:: 0.8

    :param value: IBAN string to validate
    """
    return pattern.match(value) and modcheck(value)



import re

from .utils import validator

pattern = re.compile(r'^(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$')


@validator
def mac_address(value):
    """
    Return whether or not given value is a valid MAC address.

    If the value is valid MAC address this function returns ``True``,
    otherwise :class:`~validators.utils.ValidationFailure`.

    This validator is based on `WTForms MacAddress validator`_.

    .. _WTForms MacAddress validator:
       https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

    Examples::

        >>> mac_address('01:23:45:67:ab:CD')
        True

        >>> mac_address('00:00:00:00:00')
        ValidationFailure(func=mac_address, args={'value': '00:00:00:00:00'})

    .. versionadded:: 0.2

    :param value: Mac address string to validate
    """
    return pattern.match(value)



"""Extremes."""
# -*- coding: utf-8 -*-

# standard
from functools import total_ordering
from typing import Any


@total_ordering
class AbsMax:
    """An object that is greater than any other object (except itself).

    Inspired by https://pypi.python.org/pypi/Extremes.

    Examples:
        >>> from sys import maxint
        >>> AbsMax > AbsMin
        # Output: True
        >>> AbsMax > maxint
        # Output: True
        >>> AbsMax > 99999999999999999
        # Output: True

    > *New in version 0.2.0*.
    """

    def __ge__(self, other: Any):
        """GreaterThanOrEqual."""
        return other is not AbsMax


@total_ordering
class AbsMin:
    """An object that is less than any other object (except itself).

    Inspired by https://pypi.python.org/pypi/Extremes.

    Examples:
        >>> from sys import maxint
        >>> AbsMin < -maxint
        # Output: True
        >>> AbsMin < None
        # Output: True
        >>> AbsMin < ''
        # Output: True

    > *New in version 0.2.0*.
    """

    def __le__(self, other: Any):
        """LessThanOrEqual."""
        return other is not AbsMin



"""Validate Anything!"""
# -*- coding: utf-8 -*-
# from ._extremes import AbsMax, AbsMin

from .between import between
from .btc_address import btc_address
from .card import amex, card_number, diners, discover, jcb, mastercard, unionpay, visa
from .domain import domain
from .email import email
from .hashes import md5, sha1, sha224, sha256, sha512
from .i18n import fi_business_id, fi_ssn
from .iban import iban
from .ip_address import ipv4, ipv4_cidr, ipv6, ipv6_cidr
from .length import length
from .mac_address import mac_address
from .slug import slug
from .truthy import truthy
from .url import url
from .utils import ValidationFailure, validator
from .uuid import uuid

__all__ = (
    "amex",
    "between",
    "btc_address",
    "card_number",
    "diners",
    "discover",
    "domain",
    "email",
    "fi_business_id",
    "fi_ssn",
    "iban",
    "ipv4_cidr",
    "ipv4",
    "ipv6_cidr",
    "ipv6",
    "jcb",
    "length",
    "mac_address",
    "mastercard",
    # "AbsMax",
    "md5",
    # "AbsMax",
    "sha1",
    "sha224",
    "sha256",
    "sha512",
    "slug",
    "truthy",
    "unionpay",
    "url",
    "uuid",
    "ValidationFailure",
    "validator",
    "visa",
)

__version__ = "0.20.0"



import re

from validators.utils import validator

business_id_pattern = re.compile(r'^[0-9]{7}-[0-9]$')
ssn_checkmarks = '0123456789ABCDEFHJKLMNPRSTUVWXY'
ssn_pattern = re.compile(
    r"""^
    (?P<date>(0[1-9]|[1-2]\d|3[01])
    (0[1-9]|1[012])
    (\d{{2}}))
    [A+-]
    (?P<serial>(\d{{3}}))
    (?P<checksum>[{checkmarks}])$""".format(checkmarks=ssn_checkmarks),
    re.VERBOSE
)


@validator
def fi_business_id(business_id):
    """
    Validate a Finnish Business ID.

    Each company in Finland has a distinct business id. For more
    information see `Finnish Trade Register`_

    .. _Finnish Trade Register:
        http://en.wikipedia.org/wiki/Finnish_Trade_Register

    Examples::

        >>> fi_business_id('0112038-9')  # Fast Monkeys Ltd
        True

        >>> fi_business_id('1234567-8')  # Bogus ID
        ValidationFailure(func=fi_business_id, ...)

    .. versionadded:: 0.4
    .. versionchanged:: 0.5
        Method renamed from ``finnish_business_id`` to ``fi_business_id``

    :param business_id: business_id to validate
    """
    if not business_id or not re.match(business_id_pattern, business_id):
        return False
    factors = [7, 9, 10, 5, 8, 4, 2]
    numbers = map(int, business_id[:7])
    checksum = int(business_id[8])
    sum_ = sum(f * n for f, n in zip(factors, numbers))
    modulo = sum_ % 11
    return (11 - modulo == checksum) or (modulo == 0 and checksum == 0)


@validator
def fi_ssn(ssn, allow_temporal_ssn=True):
    """
    Validate a Finnish Social Security Number.

    This validator is based on `django-localflavor-fi`_.

    .. _django-localflavor-fi:
        https://github.com/django/django-localflavor-fi/

    Examples::

        >>> fi_ssn('010101-0101')
        True

        >>> fi_ssn('101010-0102')
        ValidationFailure(func=fi_ssn, args=...)

    .. versionadded:: 0.5

    :param ssn: Social Security Number to validate
    :param allow_temporal_ssn:
        Whether to accept temporal SSN numbers. Temporal SSN numbers are the
        ones where the serial is in the range [900-999]. By default temporal
        SSN numbers are valid.

    """
    if not ssn:
        return False

    result = re.match(ssn_pattern, ssn)
    if not result:
        return False
    gd = result.groupdict()
    checksum = int(gd['date'] + gd['serial'])
    return (
        int(gd['serial']) >= 2 and
        (allow_temporal_ssn or int(gd['serial']) <= 899) and
        ssn_checkmarks[checksum % len(ssn_checkmarks)] ==
        gd['checksum']
    )



"""Between."""
# -*- coding: utf-8 -*-

# standard
from typing import TypeVar, Union
from datetime import datetime

# local
from ._extremes import AbsMax, AbsMin
from .utils import validator

PossibleValueTypes = TypeVar("PossibleValueTypes", int, float, str, datetime)


@validator
def between(
    value: PossibleValueTypes,
    /,
    *,
    min_val: Union[PossibleValueTypes, AbsMin, None] = None,
    max_val: Union[PossibleValueTypes, AbsMax, None] = None,
):
    """Validate that a number is between minimum and/or maximum value.

    This will work with any comparable type, such as floats, decimals and dates
    not just integers. This validator is originally based on [WTForms-NumberRange-Validator][1].

    [1]: https://github.com/wtforms/wtforms/blob/master/src/wtforms/validators.py#L166-L220

    Examples:
        >>> from datetime import datetime
        >>> between(5, min_val=2)
        # Output: True
        >>> between(13.2, min_val=13, max_val=14)
        # Output: True
        >>> between(500, max_val=400)
        # Output: ValidationFailure(func=between, args=...)
        >>> between(
        ...     datetime(2000, 11, 11),
        ...     min_val=datetime(1999, 11, 11)
        ... )
        # Output: True

    Args:
        value:
            Value which is to be compared.
        min_val:
            The minimum required value of the number.
            If not provided, minimum value will not be checked.
        max_val:
            The maximum value of the number.
            If not provided, maximum value will not be checked.

    Returns:
        (Literal[True]):
            If `value` is in between the given conditions.
        (ValidationFailure):
            If `value` is not in between the given conditions.

    Raises:
        AssertionError: If both `min_val` and `max_val` are `None`,
            or if `min_val` is greater than `max_val`.
        TypeError: If there's a type mismatch before comparison.

    Note:
        - `PossibleValueTypes` = `TypeVar("PossibleValueTypes", int, float, str, datetime)`
        - Either one of `min_val` or `max_val` must be provided.

    > *New in version 0.2.0*.
    """
    if min_val is None and max_val is None:
        raise AssertionError("At least one of either `min_val` or `max_val` must be specified")

    if max_val is None:
        max_val = AbsMax()
    if min_val is None:
        min_val = AbsMin()

    if isinstance(min_val, AbsMin):
        if type(value) is not type(max_val):
            raise TypeError("`value` and `max_val` must be of same type")
        return min_val <= value <= max_val

    if isinstance(max_val, AbsMax):
        if type(value) is not type(min_val):
            raise TypeError("`value` and `min_val` must be of same type")
        return min_val <= value <= max_val

    if type(min_val) is type(max_val):
        if min_val > max_val:
            raise AssertionError("`min_val` cannot be more than `max_val`")
        if type(value) is not type(min_val):  # or type(max_val):
            raise TypeError("`value` and (`min_val` or `max_val`) must be of same type")
        return min_val <= value <= max_val

    raise TypeError("`value` and `min_val` and `max_val` must be of same type")



import re

from .utils import validator

user_regex = re.compile(
    # dot-atom
    r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+"
    r"(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*$"
    # quoted-string
    r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|'
    r"""\\[\001-\011\013\014\016-\177])*"$)""",
    re.IGNORECASE
)
domain_regex = re.compile(
    # domain
    r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+'
    r'(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?$)'
    # literal form, ipv4 address (SMTP 4.1.3)
    r'|^\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)'
    r'(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$',
    re.IGNORECASE)
domain_whitelist = ['localhost']


@validator
def email(value, whitelist=None):
    """
    Validate an email address.

    This validator is based on `Django's email validator`_. Returns
    ``True`` on success and :class:`~validators.utils.ValidationFailure`
    when validation fails.

    Examples::

        >>> email('someone@example.com')
        True

        >>> email('bogus@@')
        ValidationFailure(func=email, ...)

    .. _Django's email validator:
       https://github.com/django/django/blob/master/django/core/validators.py

    .. versionadded:: 0.1

    :param value: value to validate
    :param whitelist: domain names to whitelist

    :copyright: (c) Django Software Foundation and individual contributors.
    :license: BSD
    """

    if whitelist is None:
        whitelist = domain_whitelist

    if not value or '@' not in value:
        return False

    user_part, domain_part = value.rsplit('@', 1)

    if not user_regex.match(user_part):
        return False

    if len(user_part.encode("utf-8")) > 64:
        return False

    if domain_part not in whitelist and not domain_regex.match(domain_part):
        # Try for possible IDN domain-part
        try:
            domain_part = domain_part.encode('idna').decode('ascii')
            return domain_regex.match(domain_part)
        except UnicodeError:
            return False
    return True
